/*                               FILE  _INTERPO
**
**                   FUNCTIONS TO LOAD AND INTERPOLATE DATA
**                           #include "interpo.ph"
**
**
**  FUNCTIONS
**
**   int errorcode = interead(struct interp *datp, FILE *fip)
**   int errorcode = intechkord(struct interp *datp, int chk_y_order)
**                   inteorder(struct interp *datp)
**   int errorcode = intederiv2(struct interp *idap)
**   extrema range = interange(struct interp *idap)
**   bool ok       = is_intedomain(struct interp *datp, double x1, double x2)
**   double yy     = interpolate(struct interp *idap, double xx)
**   #if defined(DINTERP)
**      double yy  = dinterp(struct interp *idap, double xx)
**   #endif
**
**  The interpolation struct defines three pointers to triplets which contain
**  interpolation data.  Each triplet datum contains the following:
**     x = independent variable (abscissa)
**     y = dependent variable (ordinate)
**     z = second derivative of y wrt x as computed under the assumptions of
**              cubic spline interpolation.
**
**       The functions interpolate and possibly intederiv2 assume increasing x
**  values in the struct.
**
**
**  References:
**   R. W. Hornbeck, _Numerical Methods_,
**        (Quantum Publishers, N. Y., 1975) pp. 47-8 & 96-7.
**   Personal Notes, _Cubic Spline Interpolation_
**        (should be in same folder as source code hard copy)
**
**
**  Associations:
**
**    This file     Personal        Hornbeck      Hornbeck
**                   Notes          pp. 47-8      pp. 96-7
**
**    nn             n              n              n
**    dp[ii].x       x[i]           x[i]
**  or cp->x
**    dp[ii].y       y[i]           f(x[i])
**  or cp->y
**    dp[ii].z       g"[i]          g"(x[i])       x[i]
**  or cp->z
**    cp1->x         x[i+1]         x[i+1]
**    cp1->y         y[i+1]         f(x[i+1])
**    cp1->z         g"[i+1]        g"(x[i+1])
**    aa[ii]         a[i]                          a[i]
**    bb[ii]         b[i]                          b[i]
**    rr[ii]         r[i]                          r[i]
**    sve            t                             TEMP
**    xx             x              x
**    dx             (del)x[i]      (del)x[i]
**    dx1            (del)x[i-1]    (del)x[i-1]
**    dx2            ((del)x[i])^2
**    dxx            x - x[i]       x - x[i]
**    dxx1           x[i+1] - x     x[i+1] - x
**
**    where [] denotes a subscript in Hornbeck and the personal notes,
**          ^ denotes an exponent, and (del) denotes a single capital delta.
**
*/
#if !defined(MAIN)
   #include "std.ph"
   #include <stdio.h>
   #include <malloc.h>
   #include "math.ph"     /* ERRDBL, MAXDBL, triplet, extrema   */
   #include "interpo.ph"

   extern uchar Bufs[],   /* at least 102 chars                 */
                Press[],  /* Press any key to continue.         */
                Errors[]; /* General error return message       */
   extern int   Erri;     /* General error return value.        */
   extern uint  Acti;     /* action flag cleared by interpolate */
                          /*   to indicate an error condition.  */
#endif

intern(byte) finddatum(struct interp *, double);
/*****************************************************************************
*/


/*                             FUNCTION  interead
**
**                  READ INTERPOLATION DATA FROM A DISK FILE
**
**  The file must be opened and the pointer positioned to the first datum to be
**  read.  The independent and dependent variables are separated by white space
**  only.  Reading stops when a line with less that two real numbers is
**  encountered or when INTEMAX (defined in interpo.ph) number of data points
**  have been read.
**
**  intederiv2 should not be called from this routine since the user may want
**  to do some fiddling with the data (e.g. wavelength to 1/wavelength) before
**  intederiv2 is called.
**
**
**  RETURN VALUE:  Erri set and returned  ( <=0 indicates error )
**    -1   less than two data points found
**    -2   file not opened
**    -3   file read error
**     0   more than INTEMAX data pairs found in the file
**     1   non data line read from file (normal return)
**     2   end of file reached (unless less than two data pts)
** ----------------------------------------------------------------------------
*/
int interead
   (
   struct interp *datp,   /* storage location of data */
   FILE          *fip     /* file to read from        */
   )
{
sint numi = 0;               /* number of data points read */
uchar *fns = " (interead)."; /* name of function for error return */
uchar *formats;              /* format string for error return    */
double test1, test2;
/* ----------------------------------------------------------------------------
*/
datp->lp = datp->mp = datp->bp;
if (fip == NULL) 
   {
   Erri = -2;
   goto EXIT;
   }
do
   {
   if (fgets(Bufs, 100, fip) == NULL) /* Use fgets/sscanf rather than fscanf */
                                      /*   since fscanf treats the new line  */
      {                               /*   like any other white space        */
      Erri = (feof(fip) ? (numi<2 ? -1 : 2) : -3);
      goto EXIT;
      }
   if (sscanf(Bufs, " %lg %lg", &datp->mp->x, &datp->mp->y) == 2) numi++;
   else
      {
      Erri = (numi<2 ? -1 : 1);
      goto EXIT;
      }
   }
while (++(datp->mp) < datp->bp + INTEMAX);

   /*     INTEMAX points read.  Check to see if there are even more data.
   */
if (fgets(Bufs, 100, fip) == NULL)  Erri = (feof(fip) ? 2 : -3);
else Erri = (sscanf(Bufs, " %lg %lg", &test1, &test2) == 2 ? 0 : 1);

EXIT:
(datp->mp)--;                                               
switch (Erri)
   {
   case -3:
      formats = "Interpolation data file read error%s";
      break;
   case -2:
      formats = "Interpolation data file not opened%s";
      break;
   case -1:
      formats = "Less than two data points found%s";
      break;
   case 0:
      sprintf(Errors, "More than %i data points found in the file%s",
                   INTEMAX, fns);
   default:
      return Erri;
   }
sprintf(Errors, formats, fns);
return Erri;
}
/*                          END OF FUNCTION interead
*******************************************************************************/


/*                            FUNCTION  inteorder
**
**        ORDER INTERPOLATION DATA FOR INCREASING INDEPENDENT VARIABLE
**
**    This function assumes that the data is already in either increasing or
**  decreasing order and inverts the order if it is decreasing.  If it ever
**  happens that data may be random,  qsort may be incorporated in this
**  function without changing its interface.
**    This function is separate from, and not called by, interead since the
**  user may want to massage the data before checking or check and massage.
** ----------------------------------------------------------------------------
*/
void inteorder(struct interp *datp)
{
triplet *tp1, *tp2, svet;

tp1 = datp->bp;  tp2 = datp->mp;
if (tp1->x < tp2->x) return;

while (tp2 > tp1)
   {
   svet = *tp1;
   *tp1 = *tp2;
   *tp2 = svet;
   tp1++;  tp2--;
   }
return;
}
/*                         END OF FUNCTION inteorder
******************************************************************************/


/*                            FUNCTION  intechkord
**
**       CHECK INTERPOLATION DATA FOR MONOTONY IN x AND y IF REQUESTED
**
**    This function verifies that interpolation data is monotonic in x, which
**  it always should be.  It also checks for monotonic y if requested.
**  This is usefull in checking for typographical errors, especially in the
**  case of index data where n values tend to be very close together.
**    This function is separate from, and not called by, interead since the
**  user may want to massage the data before checking or check and massage.
**  In fact, he may even want to set up a sort for known out-of-order data.
**  However, the error numbers are designed for incorporation into interead
**  at a later date.
**
**  RETURN VALUE:  Erri set and returned.
**       -5:     x values out of order  (if x out of order, y is not checked)
**       -4:     y values out of order
**        0:     data is monotonic
** ----------------------------------------------------------------------------
*/
int intechkord
   (
   struct interp *datp,   /* data base to check         */
   int cyi                /* cyi = 1 -> check y values. */
   )                      /*       0 -> no y check      */
{      
triplet *tp1, *tp2, *tp;
int ordi;                  /* 1 for increasing, -1 for decreasing */
tp1 = datp->bp;  tp2 = datp->mp;

ordi = (tp2->x >= tp1->x ? 1 : -1);
for (tp = tp1;  tp < tp2;  tp++)
   {
   if (0. >= ordi * (tp[1].x-tp[0].x))    /* x values cannot repeat */
      {
      sprintf(Errors, "x values in data file not monotonic:  %lg follows %lg",
                         tp[1].x, tp[0].x);
      return (Erri = -5);
      }
   }

if (cyi)
   {
   ordi = (tp2->y >= tp1->y ? 1 : -1);
   for (tp = tp1;  tp < tp2;  tp++)
      {
      if (0. > ordi * (tp[1].y-tp[0].y))  /* ok for y values to repeat */
         {
         sprintf(Errors, "y values in data file not monotonic:  %lg follows %lg",
                            tp[1].y, tp[0].y);
         return (Erri = -4);
         }
      }
   }
return (Erri = 0);
}
/*                         END OF FUNCTION intechkord
******************************************************************************/


/*                          FUNCTION  is_intedomain
**
**         DETERMINE IF THE INPUTS ARE WITHIN THE DOMAIN OF THE DATA
**
**  The interpolation data must be ordered (increasing independent variable)
**  prior to calling this function, but the x1 and x2 inputs do not.
**
**  Return true if both inputs are in the domain.  Otherwise false.
** ---------------------------------------------------------------------------
*/
bool is_intedomain
   (
   struct interp *datp, /* Interpolation data pointer   */
   double         x1,   /* Independent variables        */
   double         x2
   )
{
double min, max;

min = datp->bp->x;
max = datp->mp->x;
if (x1 >= min && x1 <= max && x2 >= min && x2 <= max) return (bool)1;
return (bool)0;
}
/*                       END OF FUNCTION is_intedomain
******************************************************************************/


/*                            FUNCTION  intederiv2
**
**       FIND SECOND DERIVATIVES FOR USE IN CUBIC SPLINE INTERPOLATION
**
**     This function was called intederiv before a function to find first
**  derivateves was contemplated.
**
**  RETURN VALUE:  Erri set and returned.
**   -1 -> failure  (Not enough memory      )
**    0 -> failure  (No or too few data read)
**    1 -> success
** ----------------------------------------------------------------------------
*/
#pragma check_stack(on)  /* Alloca used in this function */

int intederiv2(struct interp *idap)
             /* interpolation data */
{
double  *aa, *rr, dx, dx1, sve;
triplet *dp;
sint     ii, nn;
/* ----------------------------------------------------------------------------
*/
dp = idap->bp;
if (idap->mp <= dp)                /* Includes the non-initialized */
   {                               /*    case of mp = NULL (PPP?). */
   sprintf(Errors, "No or too few data read for interpolation (intederiv2).");
   return (Erri = 0);
   }
nn = idap->mp - dp;    /* Don't do this first since mp - bp may be */
                       /*    too large for an int if mp = NULL.    */
ii = nn + 1;
aa = alloca(2 * ii * sizeof(double));
if (NULL == aa) 
   {
   sprintf(Errors, "Not enough stack memory to allocate (intederiv2).");
   return (Erri = -1);
   }
rr = aa + ii;


/*         MODIFIED TRIDIAGONAL SOLVE FOR SECOND DERIVATIVES dp[ii].z
*/
aa[nn] = *aa = 0.;
rr[nn] = *rr = 0.;

for(ii = nn-1;  ii;  ii--)
   {
   dx  = dp[ii+1].x - dp[ii].x;
   dx1 = dp[ii].x   - dp[ii-1].x;
   aa[ii] = dx1/dx;
   rr[ii] = 6.*((dp[ii+1].y-dp[ii].y)/(dx*dx) - (dp[ii].y-dp[ii-1].y)/(dx*dx1));

   sve = 1./(2.0*(dx+dx1)/dx - aa[ii+1]);
   aa[ii] *= sve;
   rr[ii] = (rr[ii] - rr[ii+1])*sve;
   }

dp[0].z = 0.;
for (ii = 1;  ii <= nn;  ii++)   dp[ii].z = rr[ii] - aa[ii]*dp[ii-1].z;

return (Erri = 1);
}
/*                         END OF FUNCTION intederiv2
******************************************************************************/


/*                            FUNCTION  interange
**  
**                     FIND THE RANGE OF ORDINATE VALUES
**                            IN AN interp STRUCT
**
**  RETURN VALUE:
**     extrama type (.min & .max typedef in math.ph)
**     giving the range of ordinate (y) data values.
**
**     This function finds the max and min values of the ordinate (y) data.
**  The range returned may not be the full range which would be found by 
**  stepping through the entire domain with infinitesimal step size, but
**  should be a good approximation thereto.
** ---------------------------------------------------------------------------
*/
extrema interange(struct interp *idap)
{
extrema range;
triplet *cp;
double  yy;
/* ---------------------------------------------------------------------------
*/
range.min = MAXDBL;
range.max = -MAXDBL;

for(cp = idap->bp;  cp <= idap->mp;  cp++)
   {
   yy = cp->y;
   if (range.min > yy)  range.min = yy;
   if (range.max < yy)  range.max = yy;
   }

return range;
}
/*                         END OF FUNCTION interange
*****************************************************************************/


/*                             FUNCTION  finddatum
**
**                     FIND THE DATUM IN AN interp STRUCT
**                      LYING JUST BELOW A GIVEN ABSCISSA
**
**  This routine was part of interpolate until dinterp was written and
**  it was found to be needed twice.
**
**  RETURN VALUE:
**     The byte returned by finddatum indicates whether the datum was found:
**            0 -> found datum
**            1 -> domain error
**            2 -> bad or uninitialized data
**     If the datum was found, idap->lp (where idap points to the interpolation 
**     data) is updated to point to the datum just below the given abscissa.
**     To facilitate plotting, Acti is set if successfull and cleared on error.
**     Errors IS SET ON DOMAIN ERROR IFF Acti==1 WHEN interpolate IS CALLED:
**     When used for plotting, there may be scores of domain errors before the
**     valid domain is reached.  Calling sprintf each time will waste time.
** ---------------------------------------------------------------------------
*/
intern(byte) finddatum
   (                                                             
   struct interp *idap,  /* interpolation data - must be ordered (increasing */
                         /*    abscissa values, see function inteorder).     */
   double xx             /* abscissa (x value) to find. */
   )
{
triplet *mp,  /* pointer to last (max) datum */
        *bp,  /* base pointer to data        */
        *cp;  /* pointer to current datum    */
/* ---------------------------------------------------------------------------
*/
mp = idap->mp;  cp = idap->lp;  bp = idap->bp;

if (mp <= bp)
   {
   #if defined(DBG)
      sprintf(Errors, "Bad or uninitialized data (finddatum).  %s", Press);
   #else
      sprintf(Errors, "Bad or uninitialized data.  %s", Press);
   #endif
   Acti = 0;
   return 2;    /* Includes the non-initialized */
   }            /*    case of mp = NULL (PPP?). */


/*                      MOVE cp TO POINT TO CORRECT DATA
*/
if (xx < cp->x)         /* xx smaller than previous abscissa -> decrease */
   {                    /*    data pointer until the situation reverses. */
   do {                 /* Expect that in usual (material index) case,   */
      cp--;             /*    xx will get smaller on successive calls.   */
      if (cp < bp)
         {
         idap->lp = bp;
         goto DOMAINERR;
         }
      }
   while (xx < cp->x);
   idap->lp = cp;
   }
else                    /* xx larger than previous abscissa -> increase  */
   {                    /*    data pointer until the situation reverses. */
   do {
      cp++;
      if (cp > mp)
         {
         idap->lp = mp - 1;
         goto DOMAINERR;
         }
      }
   while (xx > cp->x);
   idap->lp = cp - 1;
   }
Acti = 1;
return 0;

DOMAINERR:
if (Acti)
   {
   #if defined(DBG)
      sprintf(Errors, "Abscissa out of domain (finddatum).  %s", Press);
   #else
      sprintf(Errors, "Abscissa out of domain.  %s", Press);
   #endif
   Acti = 0;
   }
return 1;
}
/*                         END OF FUNCTION finddatum
******************************************************************************/


/*                           FUNCTION  interpolate
**
**            INTERPOLATE AN ORDINATE (y) FOR A GIVEN ABSCISSA (x)
**
**  See function finddatum for a discussion of the various return conditions:
**      Acti, Errors, etc.
**
**  RETURN VALUE:
**     Interpolated ordinate (normal return).
**     0. is returned on domain error, ERRDBL is returned on other errors.
** ---------------------------------------------------------------------------
*/
double interpolate
   (
   struct interp *idap, /* interpolation data - must be ordered and 2nd */
                        /*    derivatives must already be calculated on */
                        /*    idap via a prior call to intederiv2.      */
   double xx            /* abscissa (x value, independent variable)     */
   )
{
triplet *cp,  /* pointer to datum just below xx (x[i] in references)   */
        *cp1; /* pointer to datum just above xx (x[i+1] in references) */

double  dx, dx2, dxx, dxx1;
/* ---------------------------------------------------------------------------
*/
switch (finddatum(idap, xx))
   {
   case 0:  /* No errors:  idap-lp positioned just below xx */
            /*             INTERPOLATE                      */

      cp   = idap->lp;               cp1  = cp + 1;

      dx   = cp1->x - cp->x;         dxx  = xx     - cp->x;
      dx2  = dx * dx;                dxx1 = cp1->x - xx;   

      return (    dxx1 * ( (dxx1*dxx1-dx2)*cp->z /6. + cp->y  )
                + dxx  * ( (dxx*dxx - dx2)*cp1->z/6. + cp1->y )
             ) / dx;             /* &&& also try integer 6 */
                                 /* Int promotes to double in multiplication */
                                 /* Still don't know about division.         */
   case 1:  /* Domain Error:  Errors and Acti set by finddatum */
      return 0.;

   case 2:  /* Data Error:  Errors and Acti set by findatum    */
      return ERRDBL;
   }
}
/*                        END OF FUNCTION interpolate
*****************************************************************************/

#if defined(DINTERP)
/*                             FUNCTION  dinterp
**
**                  FIND THE DERIVATIVE dy/dx FOR A GIVEN x
**                          FROM INTERPOLATION DATA
**  
**  I considered finding the first derivative only at data points then using
**  these as new ordinates, applying a second interpolation to find dx/dy for 
**  an arbitrary x.  Instead, this function is a direct differentiation of the
**  function interpolate.  I see no reason to prefer one to the other in terms
**  of nearness to reality.
**  
**  See function finddatum for a discussion of the various return conditions:
**      Acti, Errors, etc.
**
**  RETURN VALUE:
**     Interpolated dy/dx (normal return).
**     0. is returned on domain error, ERRDBL is returned on other errors.
** ---------------------------------------------------------------------------
*/
double dinterp
   (
   struct interp *idap, /* interpolation data - must be ordered and 2nd */
                        /*    derivatives must already be calculated on */
                        /*    idap via a prior call to intederiv2.      */
   double xx            /* abscissa (x value, independent variable)     */
   )
{
triplet *cp,  /* pointer to datum just below xx (x[i] in references)   */
        *cp1; /* pointer to datum just above xx (x[i+1] in references) */

double  dx, dx2, dxx, dxx1;
/* ---------------------------------------------------------------------------
*/
switch (finddatum(idap, xx))
   {
   case 0:  /* No errors:  idap-lp positioned just below xx */
            /*             INTERPOLATE                      */

      cp   = idap->lp;               cp1  = cp + 1;

      dx   = cp1->x - cp->x;         dxx  = xx     - cp->x;
      dx2  = dx * dx;                dxx1 = cp1->x - xx;   

      return (  ( (dx2-3*dxx1*dxx1)*cp->z + (3*dxx*dxx-dx2)*cp1->z ) / 6.
                - cp->y + cp1->y
             ) / dx;                               /* &&& also try integer 6 */

   case 1:  /* Domain Error:  Errors and Acti set by finddatum */
      return 0.;

   case 2:  /* Data Error:  Errors and Acti set by findatum    */
      return ERRDBL;
   }
}
/*                          END OF FUNCTION dinterp
*****************************************************************************/
#endif
