/*                              FILE  _RMATRIX.C
**
**       FUNCTIONS FOR HANLING REAL MATRICIES WITH ARBITRARY DIMENSIONS
**                           #include "rmatrix.ph"
**
** TTT  Should be optimized for time.
**
** Functions:
**   double *destp = mcopy(double *destp, double *srcp, int nrowi, ncoli)
**   bool   succi  = minvert(double *mtxp, int dimi)
**   double *resp  = mmul(double *resp, double *map, double *mbp,
**                            int nrowi, int nii, int ncoli)
**   void            mprint(FILE *fp,  double *mp,  int nrowi,  int ncoli)
*/
#if !defined(MAIN)
   #include "std.ph"
   #include <stdio.h>
   #include <malloc.h>
   #include <memory.h>
   #include "rmatrix.ph"
   double cdecl fabs(double);

   extern uchar   Errors[]; /* Error message buffer                   */
   #if defined(MXPRN)
      extern int  Widi,     /* Width of double output fields (> Digi) */
                  Digi;     /* Number of digits in double output      */
   #endif
#endif

intern(void) swapneg(double *, double *);

#define D2P(ii, jj)    (mtxp + dimi*ii + jj)
#define  D2(ii, jj)  (*(mtxp + dimi*ii + jj))
/******************************************************************************
*/

#if defined(MXPRN)
/*                              FUNCTION  mprint
**  
**                              OUTPUT A MATRIX
**  --------------------------------------------------------------------------
*/
void mprint
   (
   FILE   *fp,    /* File to print to  */
   double *mp,    /* Pointer to matrix */
   int     nrowi, /* Number of rows    */
   int     ncoli  /* Number of columns */
   )
{
int ii, jj;
/*  --------------------------------------------------------------------------
*/
for (ii = nrowi;  ii > 0;  ii--)
   {
   fprintf(fp, "\n");
   for (jj = ncoli; jj > 0; jj--)  fprintf(fp, "%*.*lg  ", Widi, Digi, *(mp++));
   }
fprintf(fp, "\n");
}
/*                           END OF FUNCTION mprint
******************************************************************************/
#endif


/*                              FUNCTION  mcopy
**
**                               COPY A MATRIX
** GGG
** ----------------------------------------------------------------------------
*/
double *mcopy
   (
   double *dest,  /* destination pointer     */
   double *src,   /* source pointer          */
   int    nrowi,  /* number of rows          */
   int    ncoli   /* number of columns       */
   )

    /*  Notes:  1.  dest and src should not overlap.
    **          2.  Only the product nrowi * ncoli is significant.
    */
{
return memcpy(dest, src, nrowi*ncoli*sizeof(double));
}
/*                           END OF FUNCTION mcopy
******************************************************************************/


#pragma check_stack(on)  /* Usefull if alloca is used */


/*                             FUNCTION  minvert
**
**                              INVERT A MATRIX
**
** No globals except for Errors
** ----------------------------------------------------------------------------
*/
bool minvert
   (
   double *mtxp,  /* pointer to matrix to be inverted  */
   int    dimi    /* dimension of matrix (dimi x dimi) */
   )
   
{
   /* Return value:  1 -> successfull inversion
   **                0 -> matrix could not be inverted
   ** Inverted matrix is placed at mtxp if sussesful.
   ** Failure message placed in Errors and *mtxp is undefined if failed.
   */
int    sdimi, iii, jji, kki, *ikp, *jkp;
double dt, mx;
/* ----------------------------------------------------------------------------
*/
#if defined(DBG)
   if (1 > dimi)
      {
      sprintf(Errors, "Error:  Matrix dimension less than 1 (minvert).");
      return 0;
      }
   if (NULL == mtxp)
      {
      sprintf(Errors, "Error:  Matrix pointer is NULL (minvert).");
      return 0;
      }
#endif


if (1 == dimi)     /* 1 x 1 matrix not handled properly by general algorithm */
   {
   if (0. == *mtxp) goto NOINVERT;
   else
      {
      *mtxp = 1./(*mtxp);
      return 1;
      }
   }

sdimi  = dimi*sizeof(int);
ikp = alloca(sdimi);
jkp = alloca(sdimi);
if (NULL == ikp || NULL == jkp) goto NOINVERT;

dt = 1.;
for (kki = 0;  kki < dimi;  kki++)
   {
   mx = 0.;
   do {
      do {
         for (iii = kki;  iii < dimi;  iii++)
         for (jji = kki;  jji < dimi;  jji++)
            {
            if ( fabs(mx) > fabs(D2(iii, jji)) )  continue;
            mx = D2(iii, jji);        /* Since mx starts at 0 for each ikp  */
            *ikp = iii;               /* and jkp, they will alway be set to */
            *jkp = jji;               /* something here (no initiallization */
            }                         /* is needed).                        */

         if (mx == 0.) goto NOINVERT; /* INVERSION FAILURE */
         iii = *ikp;
         }
      while (iii < kki);
      if (iii != kki)
         for (jji = 0;  jji < dimi;  jji++)  swapneg(D2P(kki,jji), D2P(iii,jji));
      jji = *jkp;
      }
   while (jji < kki);
   if (jji != kki)
      for (iii = 0;  iii < dimi;  iii++)  swapneg(D2P(iii,kki), D2P(iii,jji));

   for (iii = 0;  iii < dimi;  iii++)
      if (iii != kki)  D2(iii, kki) = -D2(iii, kki) / mx;

   for (iii = 0;  iii < dimi;  iii++)
      for (jji = 0;  jji < dimi;  jji++)
         {
         if (iii == kki || jji == kki)  continue;
         D2(iii, jji) = D2(iii, jji) + D2(iii, kki) * D2(kki, jji);
         }
   for (jji = 0;  jji < dimi;  jji++)
      if (jji != kki)  D2(kki, jji) = D2(kki, jji) / mx;
   D2(kki, kki) = 1. / mx;
   dt = dt*mx;
   ikp++;  jkp++;
}                                                              /* next kki */

for (kki = dimi - 1;  kki >= 0;  kki--)
   {
   jji = *(--ikp);
   if (jji > kki)
      for (iii = 0;  iii < dimi;  iii++)  swapneg(D2P(iii,jji), D2P(iii,kki));

   iii = *(--jkp);
   if (iii > kki)
      for (jji = 0;  jji < dimi;  jji++)  swapneg(D2P(iii,jji), D2P(kki,jji));

   }
return 1;


/*                                ERROR RETURN
*/
NOINVERT:   
   sprintf(Errors, "Matrix inversion failure.");
   #if defined(DBG)
      sprintf(Errors+24, " (minvert).");
   #endif
   return 0;
}
/*                          END OF FUNCTION minvert
******************************************************************************/


/*                               FUNCTION  mmul
**
**                         MULTIPLY TWO REAL MATRICES
**
** No globals except for Errors
** ----------------------------------------------------------------------------
*/
double * mmul
   (
   double *resp,  /* result matrix of dimension nrowi x ncoli */
   double *map,   /* first matrix of dimension nrowi x nii    */
   double *mbp,   /* second matrix of dimension nii x ncoli   */
   int    nrowi,
   int    nii,
   int    ncoli
   )
      /*  map and mbp may overlap.
      **  resp may not overlap with either map or mbp
      **     UNLESS it is equal to one of them.
      **  
      **  Return resp, where *resp = *map x *mbp.
      **  NULL is returned and message placed in Errors if there is not enough
      **     stack space for alloca (only occurs if the result is to be placed
      **     in one of the source matrices).
      */
{                                                        
double  sum,
        *arowp,  /* current row in matrix a          */
        *bcolp,  /* current col in matrix b          */
        *larowp, /* one past last row in matrix a    */
        *lbcolp, /* one past last col in matrix b    */
        *frp,    /* first element in result matrix   */
        *crp,    /* current element in result matrix */
        *cap,    /* current element in matrix a      */
        *cbp;    /* current element in matrix b      */

size_t  ressize; /* size of result vector   */
int     jj;
/* ----------------------------------------------------------------------------
*/
#if defined(DBG)
   if (1 > nrowi || 1 > nii || 1 > ncoli)
      {
      sprintf(Errors, "Error:  Matrix dimension less than 1 (mmul).");
      return NULL;
      }
   if (NULL == resp || NULL == map || NULL == mbp)
      {
      sprintf(Errors, "Error:  Matrix pointer is NULL (mmul).");
      return NULL;
      }
#endif


ressize = nrowi * ncoli * sizeof(double);
if (resp == map || resp == mbp)   frp = alloca(ressize);
else                              frp = resp;
if (NULL == frp)
   {
   sprintf(Errors, "Not enough stack memory for alloca.");
   #if defined(DBG)
      sprintf(Errors+34, " (mmul).");
   #endif
   return NULL;
   }

larowp = map + nrowi*nii;
lbcolp = mbp + ncoli;

crp = frp;
for (arowp = map;  arowp < larowp;  arowp += nii)   /* See "counter vs. */
   {                                                /*   pointer loops" */
   for (bcolp = mbp;  bcolp < lbcolp;  bcolp++)     /*   in STDPH.DOC   */
      {
      cap = arowp;
      cbp = bcolp;
      sum = 0.;
      for (jj = nii; jj > 0; jj--)
         {
         sum += (*cap) * (*cbp);
         cap ++;
         cbp += ncoli;
         }
      *(crp++) = sum;
      }
   }

if (resp == map || resp == mbp)  memcpy(resp, frp, ressize);
return resp;
}
/*                           END OF FUNCTION mmul
******************************************************************************/


/*                             FUNCTION  swapneg
**
**                 SWAP TWO DOUBLE VALUES WHILE NEGATING ONE
**  GGG
** ----------------------------------------------------------------------------
*/
intern(void) swapneg(double *ap,  double *bp)
{
double hold = *ap;
*ap = *bp;  *bp = -hold;
}
/*                          END OF FUNCTION swapneg
******************************************************************************/
