/*                              PROGRAM  CUSPLINE
**  
**                  CALCULATE AND REPORT CUBIC COEFFICIENTS
**                    FROM CUBIC SPLINE INTERPOLATION DATA
**  
**                   Must be linked with _INTERPO
**                   Link to _RMATRIX if COEFFS defined
**
**  Command Line Arguments:
**     If no arguments are given, the user is prompted for the interpolation
**  data file to read from.  Otherwise, the first argument is taken to be the
**  read file. The second argument, if given, is the name of the file to output
**  the cubic coefficients.  If no file is given, cubic coefficients are not
**  generated.  In either case, the user will be prompted for abscissa values
**  from which ordinate values will be interpolated and the pair output to
**  stdout.
**
** #if defined(COEFFS)
**  Optional Output:
**     The file containing the cubic coefficients has one line per interval.
**  The first number in each line is one endpoint of the interval covered 
**  (x.sub.n), while the first number of the next line is the other endpoint
**  (x.sub.n+1).  The remaining four numbers on each line are the cubic
**  coefficients with the zero order coeffiecint (constant term) first and
**  the third order coefficient last.
**     Note that COEFFS must be defined for compilation AND a filename given
**  as the second argument for cubic coefficients to be calculated
** #endif
**
**  Standard Output:
**     Whether or not cubic coefficients are calculated, the user will be
**  propted for absissas (independent variable values).  The prompt is sent
**  to stderr and the input received from stdin.  The entered absissa is then 
**  echoed to stdout along with an interpolated ordinate.  This feature can be 
**  used both as a check on the coefficient calculation or as a stand alone 
**  interpolator. 
**
**  Reference:
**     Handwritten notes "Calculating Cubic Coefficients from Cubic Spline Data"
**  ---------------------------------------------------------------------------
*/
#define MAIN
#define COEFFS            /* Need not define for a stand-alone interpolator */

#include "std.ph"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>   /* for exit, malloc */
#include "interpo.ph"
#if defined(COEFFS)
   #include "rmatrix.ph"
#endif

intern(void) readnext(FILE *);  intern(void) quit(uchar *);

struct interp Idata; /* Interpolation data - &&& See if malloc saves file space */
double Mp[17],       /* 4x4 matrix - see reference                           */
       Vp[5];        /* column vector - see reference                        */

FILE *Rfp;           /* Read file containing the interpolation data.         */
uchar  Fns[50],      /* Read file name.                                      */
       Bufs[110],    /* General text buffer.                                 */
       Errors[100];  /* Error message buffer used by interpolation functions */
uchar  Press[] = ""; /* Referenced by _INTERPO                               */
uint   Acti;         /* Referenced by _INTERPO                               */

#if defined(COEFFS)
   FILE    *Wfp;     /* Write file containing the cubic coefficients.    */
   triplet  Tn,      /* First data point defining the current interval.  */
            Tn1,     /* Second data point defining the current interval. */
           *Tp;
#endif
/* ****************************************************************************
*/  

void cdecl main(argc, argv)
   int argc;
   uchar *argv[];
{
   int cci;
   double xx;
/*  ---------------------------------------------------------------------------
**  
**
**                     OPEN INPUT INTERPOATION DATA FILE
*/
if (argc == 1)
   do {
      fprintf(stderr, "File name? ");
      cci = scanf("%s", Fns);
      }
   while ((cci!=1)||strlen(Fns)==0);
else
  strcpy(Fns, argv[1]);


Rfp = fopen(Fns, "r");  /* r->read */
if (Rfp == NULL)
   {
   fprintf(stderr, "No file %s.", Fns);
   exit(0);
   }


/*                    READ AND PROCESS INTERPOLATION DATA
*/
readnext(Rfp);                               /* Find the first @ line. */
if (interead(&Idata, Rfp) < 0) quit(Errors);
    inteorder(&Idata);
if (intederiv(&Idata) < 0) quit(Errors);


/*                   GENERATE CUBIC COEFFICIENTS IF DESIRED
*/
#if defined(COEFFS)

if (argc >= 3)
   {
   double xn, x1;

   #if 0
      Wfp = fopen(argv[2], "r");  /* r->read */
      if (Wfp != NULL)
         {
         fprintf(stderr, "File %s already exits.", argv[2]);
         exit(0);
         }
   #endif
   Wfp = fopen(argv[2], "a");  /* a->append */
   if (Wfp == NULL)
      {
      fprintf(stderr, "Error in creating file %s.", argv[2]);
      exit(0);
      }

   Tp = Idata.bp;
   while (Tp < Idata.mp)
      {
      Tn  = *Tp;       xn = Tn.x;
      Tn1 = *(++Tp);   x1 = Tn1.x;
      Mp[0] = 1.;  Mp[1] = xn;  Mp[2] = xn*xn;  Mp[3] = xn*xn*xn;
      Mp[4] = 1.;  Mp[5] = x1;  Mp[6] = x1*x1;  Mp[7] = x1*x1*x1;
      Mp[8] = 0.;  Mp[9] = 0.;  Mp[10]= 2.;     Mp[11]= 6*xn;
      Mp[12]= 0.;  Mp[13]= 0.;  Mp[14]= 2.;     Mp[15]= 6*x1;

      Vp[0] = Tn.y;
      Vp[1] = Tn1.y;
      Vp[2] = Tn.z;
      Vp[3] = Tn1.z;

      if (!minvert(Mp,4)) quit("Matrix could not be inverted.");
      mmul(Vp, Mp, Vp, 4, 4, 1);

      fprintf(Wfp, "%14.7lg, %14.7lg, %14.7lg, %14.7lg, %14.7lg\n",
                   Tn.x,     Vp[0],   Vp[1],   Vp[2],   Vp[3]      );
      }
   }

#endif


/*                       INTERPOLATE USER INPUT POINTS
*/
fcloseall();                /* Close first in case ^C does not close files. */

fprintf(stdout, "\n      Interpolated Points\n");
fprintf(stderr, "\nEnter absissa values, Ctrl-Z to end.\n");

for(;;)
   {
   cci = scanf(" %lg", &xx);
   if (EOF==cci || 0==cci) break;
   fprintf(stderr, "              ");            /* For screen printing, get */
                                                 /* out from under the input */
   fprintf(stdout, "%-14.7lg  %-14.7lg\n", xx, interpolate(&Idata, xx));
   }

exit(0);
}
/*                           END OF PROGRAM CUSPLINE
*****************************************************************************/


/*                             FUNCTION  readnext
**
**                            FIND THE NEXT @ LINE
** ---------------------------------------------------------------------------
*/
intern(void) readnext(FILE *filep)
{
do
   {
   if (fgets(Bufs, 100, filep) == NULL)
      {
      if (feof(filep))
         {
         fprintf(stderr, "\n@ marker not found in file %s.", Fns);
         fprintf(stderr, "\n@ must be the first character in the line "
                         "preceeding data.");
         }
      else
         {
         fprintf(stderr, "\nFile read error.");
         }
      exit(1);
      }
   }
while ('@' != *Bufs);
return;
}
/*                          END OF FUNCTION readnext
******************************************************************************/


/*                               FUNCTION  quit
*/
intern(void) quit(uchar *msg)
   {
   fprintf(stderr, "%s", msg);
   exit(1);
   }
/*                            END OF FUNCTION quit
******************************************************************************/

